from os import system as run, path as ospath
from pathlib import Path
from rich.panel import Panel
from rich.table import Table
from rich import print
from rich.console import Console
from typing import List, Optional

err_console = Console(stderr=True)

from compose_companion.configurations import Configurations


def container_up(container: str, recreate: bool, build: bool):
    scripts = Configurations().scripts
    compose_file = Configurations().compose_file

    if scripts.get("x-before-up", {}).get(container):
        for script in scripts["x-before-up"]["before-each"]:
            run(script)
        for script in scripts["x-before-up"][container]:
            run(script)

    run(
        f"docker compose -f {compose_file} up -d --remove-orphans {container} --force-recreate={recreate} --build={build}"
    )

    if scripts.get("x-after-up", {}).get(container):
        for script in scripts["x-after-up"]["after-each"]:
            run(script)
        for script in scripts["x-after-up"][container]:
            run(script)


def container_down(container: str, volumes: bool):
    scripts = Configurations().scripts
    compose_file = Configurations().compose_file

    if scripts.get("x-before-down", {}).get(container):
        for script in scripts["x-before-down"][container]:
            run(script)

    run(f"docker compose -f {compose_file} rm -sf -v={volumes} {container}")

    if scripts.get("x-after-down", {}).get(container):
        for script in scripts["x-after-down"][container]:
            run(script)


def container_create(container: str):
    compose_file = Configurations().compose_file
    run(f"docker compose -f {compose_file} create {container}")


def container_stop(container: str):
    compose_file = Configurations().compose_file
    run(f"docker compose -f {compose_file} stop {container}")


def container_pause(container: str):
    compose_file = Configurations().compose_file
    run(f"docker compose -f {compose_file} pause {container}")


def container_unpause(container: str):
    compose_file = Configurations().compose_file
    run(f"docker compose -f {compose_file} unpause {container}")


def container_start(container: str):
    compose_file = Configurations().compose_file
    run(f"docker compose -f {compose_file} start {container}")

def backup_volumes(container: str, volumes: Optional[List[str]], backup_dir: Optional[str]):
    conf_volumes = Configurations().volumes[container]

    if backup_dir:
        path_backup_dir = Path(ospath.expandvars(backup_dir.replace('~', '$HOME'))).resolve()
        if path_backup_dir.is_file():
            err_console.print("This path refers to an existing file, please pass a directory.")
            return
        path_backup_dir.mkdir(parents=True, exist_ok=True)
        back_dir = path_backup_dir
    else:
        Path('./compose-backup').resolve().mkdir(parents=True, exist_ok=True)
        back_dir = Path('./compose-backup').resolve()

    for volume in conf_volumes:
        was_picked = volume['host'] in volumes if volumes else True

        if was_picked:
            run(f"docker run --rm --volumes-from {container} -v {back_dir}:/backup ubuntu tar cvf /backup/{volume['host']}.tar {volume['container']}")
            if volumes:
                volumes.remove(volume['host'])

    for remaining in volumes:
        err_console.print(f"Volume {remaining} not declared. Skipping...")

def restore_volumes(container: str, volumes: Optional[List[str]], backup_dir: Optional[str]):
    conf_volumes = Configurations().volumes[container]

    if backup_dir:
        path_backup_dir = Path(ospath.expandvars(backup_dir.replace('~', '$HOME'))).resolve()
        if not path_backup_dir.is_dir():
            err_console.print("The given path does not exist or is not a directory.")
            return
        back_dir = path_backup_dir
    else:
        back_dir = Path('./compose-backup').resolve()
    
    for volume in conf_volumes:
        was_picked = volume['host'] in volumes if volumes else True
        backup_exists = Path(f"{back_dir}/{volume['host']}.tar").resolve().is_file()

        if was_picked and backup_exists:
            run(f"docker run --rm --volumes-from {container} -v {back_dir}:/backup ubuntu bash -c \"cd /{volume['container']} && tar xvf /backup/{volume['host']}.tar --strip 1\"")
            if volumes:
                volumes.remove(volume['host'])
        elif was_picked:
            err_console.print(f"No backup found for {volume['host']}. Skiping...")

    for remaining in volumes:
        print(err_console.print(f"Volume {remaining} not declared. Skipping..."))

def print_config():
    [config_keys, get_config] = [
        Configurations().config_keys,
        Configurations().get_config,
    ]

    grid = Table.grid(expand=True)
    grid.add_column("Key", style="cyan")
    grid.add_column("Value", style="magenta")
    for key in config_keys:
        grid.add_row(
            key,
            get_config(key),
        )

    panel = Panel(
        grid,
        title="Configurations",
        title_align="left",
        border_style="dim",
    )

    print(panel)



from os import system as run
import sys
from io import BytesIO
from typing import List, Optional

import typer
from rich import print
from rich.console import Console

from compose_companion.configurations import Configurations

from compose_companion.functions import (
    container_down,
    container_stop,
    container_start,
    container_pause,
    container_unpause,
    container_create,
    container_up,
    print_config,
    backup_volumes,
    restore_volumes
)

app = typer.Typer()
err_console = Console(stderr=True)


def autocomplete_containers(incomplete: str):
    completion = []

    try:
        stdout_bak = sys.stdout
        sys.stdout = BytesIO()
        services = Configurations().services
        sys.stdout = stdout_bak
    except:
        services = []

    for name in services:
        if name.startswith(incomplete):
            completion.append(name)

    return completion


@app.callback(invoke_without_command=True)
def main(ctx: typer.Context):
    """
    A companion for Docker Compose.
    """
    if not ctx.invoked_subcommand:
        ctx.get_help()
    else:
        Configurations()

@app.command()
def up(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
    recreate: bool = typer.Option(
        False,
        "--recreate",
        "-r",
        help="Recreates containers even if their configuration and image haven't changed. Toggles docker compose --force-recreate option.",
    ),
    build: bool = typer.Option(
        False,
        "--build",
        "-b",
        help="Build images before starting containers. Toggles docker compose --build option.",
    ),
):
    """
    Starts up targeted containers.
    If no target is defined, starts all containers.
    Calls to `docker compose up -d --remove-orphans`.
    """
    services = Configurations().services
    if not targets:
        for container in services:
            container_up(container, recreate, build)
        return
    for container in targets:
        container_up(container, recreate, build)
    return


@app.command()
def start(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
):
    """
    Starts targeted containers.
    If no target is defined, starts all containers.
    Equivalent to `docker compose start`.
    """
    services = Configurations().services
    if not targets:
        for container in services:
            container_start(container)
        return
    for container in targets:
        container_start(container)


@app.command()
def pause(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
):
    """
    Pauses targeted containers.
    If no target is defined, pauses all containers.
    Equivalent to `docker compose pause`.
    """
    services = Configurations().services
    if not targets:
        for container in services:
            container_pause(container)
        return
    for container in targets:
        container_pause(container)


@app.command()
def unpause(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
):
    """
    Unpauses targeted containers.
    If no target is defined, unpauses all containers.
    Equivalent to `docker compose unpause`.
    """
    services = Configurations().services
    if not targets:
        for container in services:
            container_unpause(container)
        return
    for container in targets:
        container_unpause(container)


@app.command()
def stop(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
):
    """
    Stops targeted containers.
    If no target is defined, stops all containers.
    Equivalent to `docker compose stop`.
    """
    services = Configurations().services
    if not targets:
        for container in services:
            container_stop(container)
        return
    for container in targets:
        container_stop(container)


@app.command()
def create(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
):
    """
    Creates targeted containers.
    If no target is defined, creates all containers.
    Equivalent to `docker compose create`.
    """
    services = Configurations().services
    if not targets:
        for container in services:
            container_create(container)
        return
    for container in targets:
        container_create(container)


@app.command()
def down(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
    remove_volumes: Optional[bool] = typer.Option(
        False,
        "-v",
        "--volumes",
        help="Remove any anonymous volumes attached to the containers. Equivalent do docker compose -v/--volumes option.",
    ),
    confirmation: bool = typer.Option(
        False,
        "-f",
        "--force",
        prompt="You're about to shut down your containers. Are you sure?",
        help="Skip confirmation prompt and run.",
    ),
):
    """
    Shuts down target containers.
    If no target is defined, shuts down all containers.
    Equivalent to `docker compose rm -sf`.
    """
    services = Configurations().services

    if not confirmation:
        return
    if not targets:
        for container in services:
            container_down(container, remove_volumes)
        return
    for container in targets:
        container_down(container, remove_volumes)


@app.command()
def logs(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
    detach: bool = typer.Option(
        False,
        "--detach",
        "-d",
        help="Print logs without following. Negates docker compose --follow/-f flag, which is used by default.",
    ),
    timestamps: bool = typer.Option(
        False,
        "--timestamps",
        "-t",
        help="Show timestamps. Toggles docker compose --timestamps/-t flag.",
    ),
):
    """
    Prints logs for target containers.
    If no target is defined, prints for all running ones.
    Equivalent to `docker compose logs -f`
    """
    compose_file = Configurations().compose_file

    if not targets:
        run(f"docker compose -f {compose_file} logs -f={not detach} -t={timestamps}")
        return
    run(
        f"docker compose -f {compose_file} logs {' '.join(targets)} -f={not detach} -t={timestamps}"
    )


@app.command()
def exec(
    target: str = typer.Argument(
        ...,
        help="The target container.",
    ),
    detach: Optional[bool] = typer.Option(
        False,
        "-d",
        "--detach",
        help="Detached mode: Run command in the background. Equivalent to docker compose -d flag.",
    ),
    command: Optional[List[str]] = typer.Argument(
        None,
        help="Command to be run. By default, runs `sh`.",
    ),
):
    """
    Runs command on the container.
    Equivalent to `docker compose exec`.
    """
    compose_file = Configurations().compose_file

    if not command:
        run(f"docker compose -f {compose_file} exec -d={detach} {target} sh")
        return
    run(
        f"docker compose -f {compose_file} exec -d={detach} {target} {' '.join(command)}"
    )


@app.command()
def ps(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
    include_stopped: Optional[bool] = typer.Option(
        False,
        "-a",
        "--all",
        help="Show all, including stopped containers. Equivalent to docker compose -a/--all option.",
    ),
):
    """
    Lists containers.
    If no target is defined, lists all containers.
    Equivalent to `docker compose ps`.
    """
    compose_file = Configurations().compose_file

    if not targets:
        run(f"docker compose -f {compose_file} ps")
        return
    run(f"docker compose -f {compose_file} ps {' '.join(targets)} -a={include_stopped}")


@app.command()
def backup(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
    volumes: Optional[List[str]] = typer.Option(
        None,
        "-v",
        "--volumes",
        help="Volumes for which to run the command. If no targets are passed, runs for all volumes.",
    ),
    backup_dir: Optional[str] = typer.Option(
        None,
        "-d",
        "--backup-dir",
        help="Directory in which to save the backup. If not specified, backup will be saved on <current_path>/compose-backup",
    ),
):
    """
    Backs up named volumes for containers.
    If no target is defined, Backs up volumes for all containers.
    """
    compose_file = Configurations().compose_file
    services = Configurations().services

    if not targets:
        for container in services:
            backup_volumes(container, volumes, backup_dir)
        return
    for container in targets:
        backup_volumes(container, volumes, backup_dir)


@app.command()
def restore(
    targets: Optional[List[str]] = typer.Argument(
        None,
        help="Containers in which to run the command. If no targets are passed, runs for all containers.",
        autocompletion=autocomplete_containers,
    ),
    volumes: Optional[List[str]] = typer.Option(
        None,
        "-v",
        "--volumes",
        help="Volumes for which to run the command. If no targets are passed, runs for all volumes.",
    ),
    backup_dir: Optional[str] = typer.Option(
        None,
        "-d",
        "--backup-dir",
        help="Directory from which to fetch the backup. If not specified, backup will be fetched from <current_path>/compose-backup",
    ),
):
    """
    Restores previously backed up named volumes for containers.
    If no target is defined, Restores volumes for all containers.
    """
    compose_file = Configurations().compose_file
    services = Configurations().services

    if not targets:
        for container in services:
            restore_volumes(container, volumes, backup_dir)
        return
    for container in targets:
        restore_volumes(container, volumes, backup_dir)


@app.command()
def config(
    key: Optional[str] = typer.Argument(
        None,
        help="Configuration key to view/set. If left blank, prints all the conf.",
    ),
    value: Optional[str] = typer.Argument(
        None,
        help="Value to set on the key. If key is passed but not value, current value for the config will be printed.",
    ),
):
    """
    View or change configurations.
    """
    conf = Configurations()

    if key and key in conf.config_keys:
        if key == "compose_file" and value:
            conf.compose_file = value
        elif key == "companion_file" and value:
            conf.companion_file = value
        print(conf.get_config(key))
        return

    if key and key not in conf.config_keys:
        err_console.print("Invalid Key. Here are the available keys:\n")

    print_config()


if __name__ == "__main__":
    app()



import configparser
from pathlib import Path
from os import path as ospath
import typer
import yaml_env_var_parser as yaml
from dotenv import load_dotenv
from rich import print
from rich.console import Console

err_console = Console(stderr=True)


load_dotenv()


class Configurations:
    _instance = None

    current_path = Path(__file__).parent.resolve()

    config_keys = ["companion_file", "compose_file"]
    config_parser = configparser.ConfigParser()
    config_file = f"{current_path}/config.cfg"

    scripts = dict()
    services = []
    volumes = dict()
    _companion_file = ""
    _compose_file = ""

    def __new__(cls):
        if cls._instance is None:
            cls._instance = super().__new__(cls)

            # initialization
            cls._load_config_file(cls._instance)
            cls._load_compose_files(cls._instance)

        return cls._instance

    def _resolve_path(value):
        return Path(ospath.expandvars(value.replace('~', '$HOME'))).resolve()

    def _save(self):
        with open(self.config_file, "w") as file:
            self.config_parser.write(file)

    def get_config(self, key: str):
        return self.config_parser["DEFAULT"][key]

    def set_config(self, key: str, value: str):
        self.config_parser["DEFAULT"][key] = value
        self._save()

    @property
    def companion_file(self) -> str:
        try:
            self._companion_file = self.get_config("companion_file")
        except KeyError:
            self.set_config("companion_file", "")
            self._companion_file = ""
        return self._companion_file

    @companion_file.setter
    def companion_file(self, value: str) -> None:
        if not value:
            self._companion_file = ""
        else:
            self._companion_file = f"{self._resolve_path(value)}"
        self.set_config("companion_file", self._companion_file)

    @property
    def compose_file(self) -> str:
        try:
            self._compose_file = self.get_config("compose_file")
        except KeyError:
            self.set_config("compose_file", "")
            self._compose_file = ""
        return self._compose_file

    @compose_file.setter
    def compose_file(self, value: str) -> None:
        if not value:
            self._compose_file = ""
        else:
            self._compose_file = f"{self._resolve_path(value)}"
        self.set_config("compose_file", self._compose_file)

    def _load_config_file(self):
        try:
            with open(self.config_file, "r") as file:
                self.config_parser.read_file(file)
        except FileNotFoundError:
            self._save()

    def _request_compose_file(self):
        local_compose_file_yaml = Path("docker-compose.yaml").resolve()
        local_compose_file_yml = Path("docker-compose.yml").resolve()

        if local_compose_file_yaml.is_file():
            self.compose_file = str(local_compose_file)
            return
        elif local_compose_file_yml.is_file():
            self.compose_file = str(local_compose_file)
            return

        compose_file_input = typer.prompt(
            "Please informe the path to your docker compose file (including filename)"
        )

        compose_file_path = self._resolve_path(compose_file_input)
        if not compose_file_path.is_file():
            err_console.print("The informed path is invalid.\n")
            return

        self.compose_file = str(compose_file_path)
        print("Your file path has been saved!\n")

    def _request_companion_file(self):
        local_companion_file_yaml = Path("compose-companion.yaml").resolve()
        local_companion_file_yml = Path("compose-companion.yml").resolve()

        if local_companion_file_yaml.is_file():
            self.companion_file = str(local_companion_file)
            return
        if local_companion_file_yml.is_file():
            self.companion_file = str(local_companion_file)
            return
        elif Path(self.compose_file).is_file():
            use_compose = typer.confirm(
                "I didn't find your companion file, but I do have your docker-compose file. Would you like to use it for your companion scripts?"
            )
            if use_compose:
                self.companion_file = self.compose_file
                return

        companion_file_input = typer.prompt(
            "Please informe the path to your companion scripts file (including filename)"
        )

        companion_file_path = self._resolve_path(companion_file_input)
        if not companion_file_path.is_file():
            err_console.print("The informed path is invalid.\n")
            return
        self.companion_file = str(companion_file_path)
        print("Your file path has been saved!\n")

    def _load_compose_files(self):
        if not self.compose_file or not Path(self.compose_file).is_file():
            self._request_compose_file()
            self._load_compose_files()
            return

        if not self.companion_file or not Path(self.companion_file).is_file():
            self._request_companion_file()
            self._load_compose_files()
            return

        try:
            with open(self.compose_file, "r") as file:
                parsed = yaml.load(file)                
                if parsed:
                    services_dict = dict(parsed).get("services", {})
                    volumes_dict = dict(parsed).get("volumes", {})

                self.services = list(services_dict.keys())
                for service_name in self.services:
                    self.volumes[service_name] = []

                volumes_list = list(volumes_dict.keys())
                for key, value in services_dict.items():
                    service_vols = value.get("volumes", [])
                    for service_vol in service_vols:
                        if service_vol.split(":")[0] in volumes_list:
                            self.volumes[key].append({'host': service_vol.split(":")[0], 'container': service_vol.split(":")[1]})

                
        except (FileNotFoundError, IsADirectoryError) as err:
            err_console.print("Docker Compose file path is invalid")
            self.compose_file = ""
            self._load_compose_files()
            return
        except ValueError as err:
            compose_file_dir = Path(self.compose_file).parent.resolve()
            dotenv_file_path = Path(f"{compose_file_dir}/.env")
            if dotenv_file_path.is_file():
                load_dotenv(dotenv_file_path)
                self._load_compose_files()
                return
        except TypeError as err:
            err_console.print(
                "Invalid syntax somewhere on your docker compose file. Please make sure you have a valid yaml file!"
            )
            err_console.print(
                "To check or change the current file path run compose config."
            )

        try:
            with open(self.companion_file, "r") as file:
                parsed = yaml.load(file)
                if parsed:
                    self.scripts = dict(parsed)
        except (FileNotFoundError, IsADirectoryError) as err:
            err_console.print("Scripts file path is invalid")
            self.companion_file = ""
            self._load_compose_files()
            return
        except ValueError as err:
            companion_file_dir = Path(self.companion_file).parent.resolve()
            dotenv_file_path = Path(f"{companion_file_dir}/.env")
            if dotenv_file_path.is_file():
                load_dotenv(dotenv_file_path)
                self._load_compose_files()
                return
        except TypeError as err:
            err_console.print(
                "Invalid syntax somewhere on your scripts file. Please make sure you have a valid yaml file!"
            )
            err_console.print(
                "To check or change the current file path run compose config."
            )



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['compose_companion']

package_data = \
{'': ['*']}

install_requires = \
['python-dotenv>=0.21.0,<0.22.0',
 'pyyaml>=6.0,<7.0',
 'rich>=12.6.0,<13.0.0',
 'typer>=0.7.0,<0.8.0',
 'yaml-env-var-parser>=1.1.1,<2.0.0']

entry_points = \
{'console_scripts': ['compose = compose_companion.cli:app']}

setup_kwargs = {
    'name': 'compose-companion',
    'version': '0.2.0',
    'description': 'A companion for Docker Compose',
    'long_description': "# Compose Companion\n\nThis is a little CLI tool created for my home server.\n\nIt aims to make it easy to configure and document scripts that should run before and/or after the server containers on docker compose go up or down.\n\n## Scrips File\n\nThe app will read the scripts from a yaml file in the following format:\n\n```yaml\n# compose-companion.yaml\n\nx-before-up:\n  sonarr:\n    - echo this will run before sonarr startup\n    - echo this too will run before sonarr startup, after the previous one\n  radarr:\n    - echo this will run before radarr startup\n    - echo this too will run before radarr startup, after the previous one\n\nx-after-up:\n  sonarr:\n    - echo this will run after sonarr startup\n    - echo this too will run after sonarr startup, after the previous one\n  radarr:\n    - echo this will run after radarr startup\n    - echo this too will run after radarr startup, after the previous one\n\nx-before-down:\n  sonarr:\n    - echo this will run before sonarr shutdown\n    - echo this too will run before sonarr shutdown, after the previous one\n  radarr:\n    - echo this will run before radarr shutdown\n    - echo this too will run before radarr shutdown, after the previous one\n\nx-after-down:\n  sonarr:\n    - echo this will run after sonarr shutdown\n    - echo this too will run after sonarr shutdown, after the previous one\n  radarr:\n    - echo this will run after radarr shutdown\n    - echo this too will run after radarr shutdown, after the previous one\n```\n\nThe container keys should match the ones from `docker-compose.yaml` file.  \nThe app will look for a file named `compose-companion.yaml` on the folder it's first run, if that's not there it'll ask you to inform the file path manually.  \nAs the top-level keys start with `x-`, you can use the `docker-compose.yaml` file itself, if you wish, and these settings will be properly ignored by docker compose.  \n\n## Commands\n\nFor a list of command, run `compose --help` or simply `compose`.  \nFor details on each command, run `compose [command] --help`.\n",
    'author': 'Natália Fonseca',
    'author_email': 'natalia@nataliafonseca.dev',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'None',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.10,<4.0',
}


setup(**setup_kwargs)
